#!/usr/bin/perl -w
#
# first2sort
# Takes a parsed PFS:First Choice file
#  (parsed with first2html) and a command line
#  argument of the fields to include.

# Argument is a quote triplet, separated by :
#   "list fields:sort fields:summary"
# - list fields is a list of fields numbers to show (like first2list.pl)
# - sort fields is a list of fields numbers to sort on (like first2list.pl)
# - summary is the number of fields (from the left) to not repeat if duplicated

# Argument is a quoted string, of field numbers,
#  separated by commas, of fields.

# Examples:
# "1,2:3:" = show fields 1 and 2, but sort by three, summarize all

# "1,3,4" = field 1, 3 ,4 (not 2)
# "-4,6"  = field 1,2,3,4,6
# etc
	
# command line argument hiding
my ($f,$s,$rep) = ("-","","") ;
if ( $#ARGV >= 0 ) {
	($f,$s,$rep) = split(":",join(":",@ARGV)."::") ;
}
@ARGV = () ;

# Need to store all the data in a hash
my %data ;

# Add a number to break ties
my $count = 100000 ;

# Copy until <TABLE> line, inclusive
while (<>) {
	if (/<TABLE/i) {
		# print <TABLE> line -- modify table attributes here, if desired
		print ;
		last ;
	} else {
		print ;
	}
}

# Copy until <TH> line - should be next one
my $fnames ;
while (<>) {
	if (m/<TH/i) {
		# Get Field names
		@fnames = m/<TH.*?>(.*?)<\/TH>/ig ;
		last ;
	} else {
		print ;
	}
}
my $fnumber =  $#fnames + 1 ;

# parse the 1st command line argument
$f =~ s/\s*-\s*/-/g ;	# remove whitespace around -
$f =~ s/\s*,\s*/,/g ;	# remove whitespace around ,
$f =~ s/\s+/,/g ;	# change whitespace to ,
$f =~ s/^,// ;		# no trailing ,
$f =~ s/,$// ;		# no leading ,
$f =~ s/^$/-/ ;		# if blank, assume - (all)
my @f = map({if (/(\d+)-(\d+)/) {($1..$2)} elsif (/-(\d+)/) {(1..$1)} elsif (/(\d+)-/) {($1..$fnumber)} elsif (/-/) {(1..$fnumber)} else {$_} }  split(",",$f));

# parse the 2nd command line argument
$s =~ s/\s*-\s*/-/g ;	# remove whitespace around -
$s =~ s/\s*,\s*/,/g ;	# remove whitespace around ,
$s =~ s/\s+/,/g ;	# change whitespace to ,
$s =~ s/^,// ;		# no trailing ,
$s =~ s/^$// ;		# no leading ,
if ( $s eq "" ) { 
	@s = @f ; # if blank, assume - (all)
} else {
	@s = map({if (/(\d+)-(\d+)/) {($1..$2)} elsif (/-(\d+)/) {(1..$1)} elsif (/(\d+)-/) {($1..$fnumber)} elsif (/-/) {(1..$fnumber)} else {$_} }  split(",",$s));
}	

# parse the 3rd command line argument
if ( $rep =~ m/(\d+)/ ) { $rep = $1 } else { $rep = $#f + 1 } 

# print Field Names
# Modify to change BGCOLORs, fonts, etc
print "<TR>" ;
print map( "<TH>$fnames[$_-1]</TH>" , @f ) ;
print "</TR>\n" ;

# loop through records
my @ofields = @fnames ; # previous fields to substitute for &nbsp;
my $tableend ;
while(<>) {
	# increment counter
	++ $count ;
	
	if ( m/<\/TABLE/i) { print "found end \n" ; $tableend=$_ ; last ;} ; # break if not a table row
	
	my @nfields = m/<TD.*?>(.*?)<\/TD>/ig ;
	
	# process the new fields if desired

	# replace &nbsp with previous values
	# this allows sorted lists with blank repeat
	# fields to be repopulated
	my @fields = map {my $o = shift(@ofields) ; if (/\&nbsp\;/) {$o} elsif (/^$/) {" "} else {$_} }  @nfields ;

	# Move new fields to old have ready for substitution
	@ofields = @fields ;

	# Do any further processing here
	
	# Now store: (use 0xFF as a field separator)
	$data{join("",map({$fields[$_-1]} @s))."".$count} = join("",map({$fields[$_-1]} @f)) ;
}
	
# Now print
my $first = 1 ; # flag first row
foreach (sort keys %data) {
	my @fields = split "", $data{$_} ;
	# Change duplicate fields to "&nbsp"
	if ( $first ) {
		$first = 0 ; # next time not first
		@ofields = @fields ; # remember these values
	 } else {
#		not first time, see about repeat suppression
		for ( my ($i,$nomatch) = (0,0) ; $i < $rep ; $i++ ) {
			if ( $nomatch ) {
	 			$ofields[$i] = $fields[$i] ; #update remembered value
	 		} elsif ( $fields[$i] eq $ofields[$i] ) {
	 			$fields[$i] = "&nbsp;" ; # suppressed marker
	 		} else {
	 			$ofields[$i] = $fields[$i] ; #update remembered value
	 			$nomatch = 1 ;
	 		}
	 	}
	 }
	 
	# Print the record, modify TD attributes (like BGCOLOR) if desired.
	 print "<TR><TD>".join("</TD><TD>",@fields)."</TD></TR>\n" ;
}		

# Final copy -- the </TABLE> line and everything else
print $tableend ;
while (<>) {
	print ;
}
