// First Choice database reading program
// HTML Table output
//
// Paul Alfille -- Dec 2002
//
// <<<USAGE>>>
// Usage firstchoice2html inputfile.fol
// Usage firstchoice2tell inputfile.fol

// <<<TARGET>>>
// Combined source first2html and first2tell
// Compiles to first2html unless FIRST2TELL defined
// alternatives: TELL FIRSTCHOICE2TELL
#ifdef  FIRSTCHOICE2TELL
	#define FIRST2TELL
#endif
#ifdef  TELL
	#define FIRST2TELL
#endif
#ifndef	FIRST2TELL
    #define	FIRST2HTML
#endif

// <<<PLATFORM>>>
// Note currently tested under
//	gcc and linux (default)
//	Pacific C and DOS (__PACIFIC__)
#ifdef   __PACIFIC__
	#include <unixio.h>	// read
	#include <stdlib.h>	// exit
	#define  O_RDONLY	0
#else
	#include <unistd.h> // for read and write
	#include <sys/types.h>
	#include <sys/stat.h>
	#include <fcntl.h>
#endif
#include <string.h> // for memcpy
#include <stdio.h> // for sscanf
#include <time.h> // ctime


#define BYTE(c,o)		( c[o] & 0xff )
#define BYTE2(c,o)	( BYTE(c,o) + 256 * BYTE(c,o+1) )

// Globals
#define BLOCK	0x80
int	db_fieldnameblock ; // first digit of file, seems to be 8 or 9
int	db_blocks, db_blocks2 ; // file size in 128-byte blocks (plus 128)
int	db_records ; // actuual number of good records
int	db_fields ; // Number of fields defined in header
//int	db_fieldlength ; // Length of fields defined in header
//int	db_unknown ; // Unknown field  - byte 26,27
//int	db_bonus ; // ?

int fd ; // file descriptor for input file

char ReadBlock[BLOCK] ;
char Tag[] = "GERBILDB3   " ;

typedef int BOOL ;


#ifdef FIRST2TELL
    // Field list for storing summary data
	#define MAXFIELDS  100
	int fldlen[MAXFIELDS] ;

    // hold position in file for error messages
	int recpos = 0 ;
	int rb(void) { recpos+=128 ; return read( fd , ReadBlock , BLOCK ) ; }
	#define readblock	rb()
	
    // print error messages
	#define	printerr1(x)		printf( x )
	#define	printerr3(x,a1,a2)	printf(x,a1,a2)
#else
	#define	readblock	read( fd , ReadBlock , BLOCK )
	#define	printerr1(x)	;
	#define	printerr3(x,a1,a2)	;

#endif

// print out a HTML-friendly character.
void HTMLchar( char c ) {
	switch (c) {
		case 0x22:	printf("&quot;") ; 	return ;
		case '>':	printf("&gt;") ; 	return ;
		case '<':	printf("&lt;") ; 	return ;
		case '&':	printf("&amp;") ; 	return ;
		case 0x0D:	printf(" ") ; 		return ;
	}
	if ( c >= 0x7f ) {
		printf ( "&#%d;",c ) ;
	} else if ( c > 0x10 ) {
		printf( "%c",c ) ;
	}
	return ;
}

// HTML string char by char, trimming leading and trailing whitespace.
// NOTE: Must see a null char between invocations (static variables)
// returns 1 unless an 0x0D, then 2
int HTMLstring( char c ) {
	static int trailing_whitespace = 0 ;
	static BOOL instring = 0 ;
	int ret = 1 ;
	
	switch (c) {
		case 0x00:
			trailing_whitespace = 0 ;
			instring = 0 ;
			break ;
		case 0x0D:
			ret = 2 ; // fall through
		case ' ':
			if ( instring ) ++trailing_whitespace ;
			break ;
		default:
			instring = 1 ;
			for ( ; trailing_whitespace ; --trailing_whitespace ) HTMLchar(' ') ;
			HTMLchar(c) ;
	}
	return ret ;
}
			
#ifdef FIRST2TELL
// INFO string.
// Similar to HTMstring, but doesn't print anything. Really just to figure fields lengths.
int INFOstring( char c , int * reallen ) {
	static int trailing_whitespace ;
	static BOOL instring ;
	int ret = 1 ;
	
	switch (c) {
		case 0x00: // initialize
			trailing_whitespace = 0 ;
			instring = 0 ;
			*reallen = 0 ;
			break ;
		case 0x0D:
			ret = 2 ; // fall through
		case ' ':
			if ( instring ) ++trailing_whitespace ;
			break ;
		default:
			instring = 1 ;
			(*reallen) += trailing_whitespace+1 ;
			trailing_whitespace = 0 ;
	}
	return ret ;
}
#endif
			
int HTMLheaderstring( char c ) {
	switch ( (unsigned char) c) {
		case 0x90:
			return 1 ;
		case 0x80:
			return HTMLstring( ' ' ) ;
		default:
			return HTMLstring(c & 0x7F) ;
	}
}

// Read the Header record
// Special cases if errors or summary information wanted
void readhead( void ) {
	int i ;
	struct stat filestat ;

	if ( readblock != BLOCK ) {
		printerr1( "<P>Cannot read the file header</P>" ) ;
		exit(1) ;
	}

	if ( memcmp( ReadBlock+9 , Tag , sizeof(Tag) ) ) {
		printerr1( "<P>Missing *GERBILD3   * tag -- doubt this is really a First Choice database file</P>" ) ;
		exit(1) ;
	}

	db_fieldnameblock = BYTE( ReadBlock , 0 ) ;
	db_blocks = BYTE2( ReadBlock , 2 ) ;
	db_blocks2 = BYTE2( ReadBlock , 4 ) ;
	if (fstat( fd , &filestat )) {
		printerr1( "<P>Cannot get file information from operating system.</P>") ;
	} else {
		printerr3( "<P>Operating system claims file is %d bytes long. </P><P>Last modified: %s</P>", filestat.st_size, ctime(&(filestat.st_mtime)) ) ;
		printerr3( "<P>Number of blocks %d, predicted length %d bytes</P>", db_blocks, db_blocks*128+128 ) ;
	}
	if ( db_blocks != db_blocks2 ) printerr3( "<P>Number of blocks differ in the header: %d and %d</P>", db_blocks, db_blocks2 ) ;
	db_records = BYTE2( ReadBlock , 6 ) ;
	db_fields = BYTE2( ReadBlock , 22 ) ;
#ifdef FIRST2TELL
	printerr3( "<P>There are %d fields and %d records</P>" , db_fields , db_records) ;
	for ( i=0 ; i<db_fields ; ++i ) { if (i<MAXFIELDS) fldlen[i]=0 ; }
#endif
//	db_fieldlength = BYTE2( ReadBlock , 24 ) ;
//	db_unknown = BYTE2( ReadBlock , 26 ) ;
//	db_bonus = BYTE2( ReadBlock , 28 ) ;


	for ( i=db_fieldnameblock-1 ; i>0 ; --i ) {
		if ( readblock != BLOCK ) {
			printerr1( "<P>File ends while still in header range.</P>" ) ;
			exit(1) ;
		}
	}
}
	
void nextchar( int * o , int * b , int blocks ) {
	++(*o) ;
	if ( *o >= BLOCK ) {
		if ( readblock != BLOCK ) {
			*b = blocks+1 ;
			*o = 0 ;
			ReadBlock[0] = 0x00 ;
		} else {
			++(*b) ;
			*o = 2 ;
		}
	}
}

BOOL readrecord( void ) {
	int blocks ; // number of blocks
	int type ; // record type
	int b = 1 ; // block counter
	int o ; // char offset in block
	int fld ; // field counter
	int len ; // field length
	int reallen ;

	if ( readblock != BLOCK ) return 0 ;
	type = BYTE2( ReadBlock , 0 ) ;
	blocks = BYTE2( ReadBlock , 2 ) ;
	if ( type == 0x81 ) {
		o = 3 ;
#ifdef FIRST2HTML
		printf("<TR>") ;
		for ( fld = 0 ; fld < db_fields ; ++ fld ) {
			nextchar( &o , &b , blocks ) ;
			len = BYTE(ReadBlock,o)*256 ;
			nextchar( &o , &b , blocks ) ;
			len += BYTE(ReadBlock,o) - (fld==db_fields-1) ;
			
			printf("<TD>") ;
			while( (len>0) && (b<=blocks) ) {
				nextchar( &o , &b , blocks ) ;
				len -= HTMLstring(ReadBlock[o]) ;
			}
			HTMLstring(0) ;
			printf("</TD>") ;
		}
		printf("</TR>\n") ;
#else
		for ( fld = 0 ; fld < db_fields ; ++ fld ) {
			if ( fld>=MAXFIELDS ) break ;
			INFOstring(0,&reallen) ; // initialize
			nextchar( &o , &b , blocks ) ;
			len = BYTE(ReadBlock,o)*256 ;
			nextchar( &o , &b , blocks ) ;
			len += BYTE(ReadBlock,o) - (fld==db_fields-1) ;
			
			while( (len>0) && (b<=blocks) ) {
				nextchar( &o , &b , blocks ) ;
				len -= INFOstring(ReadBlock[o],&reallen) ;
			}
			if ( fldlen[fld]<reallen ) fldlen[fld]=reallen ; // maximum
		}
#endif
		for ( ; b<blocks ; ++b ) {
			if ( readblock < BLOCK ) return 0 ;
		}
		return ( b <= blocks ) ;
	} else if ( type == 0x82 ) {
		o = 7 ;
		printf("<TR>") ;
		for ( fld = 0 ; fld < db_fields ; ++ fld ) {
			nextchar( &o , &b , blocks ) ;
			len = BYTE(ReadBlock,o)*256 ;
			nextchar( &o , &b , blocks ) ;
			len += BYTE(ReadBlock,o) - (fld==db_fields-1) ;
			
			printf("<TH>") ;
			while( (len>0) && (b<=blocks) ) {
				nextchar( &o , &b , blocks ) ;
				len -= HTMLheaderstring(ReadBlock[o]) ;
			}
			HTMLstring(0) ;
			printf("</TH>") ;
		}
		printf("</TR>\n") ;
		for ( ; b<blocks ; ++b ) {
			if ( readblock < BLOCK ) return 0 ;
		}
		return ( b <= blocks ) ;
	} else if ( type != 0 ) {
		printerr3( "<TR><TD>Bad block type %d at byte position %d</TD></TR>" , type , recpos-128 ) ;
	}
	return 1 ;
}
	
#ifdef FIRST2TELL
void fieldlength( void ) {
	int i ;
	printf ( "<TR>" ) ;
	for ( i=0 ; i<db_fields ; ++i ) {
		printf ("<TD>");
		if ( i<MAXFIELDS) {
			printf( "%d" , fldlen[i] ) ;
		} else {
			printf ( "More than %d fields",MAXFIELDS ) ;
		}
		printf ("</TD>");
	}
	printf ( "</TR>\n" ) ;
}
#endif

int main( int argc , char * argv[] ) {
	
	if ( argc == 1 ) {
		fprintf( stderr , "Parse a PFS:First Choice (old DOS program) database into an HTML table\n\t%s filename\n", argv[0] ) ;
		exit(1) ;
	}

	fd = open( argv[1] , O_RDONLY ) ;
	if ( fd < 0 ) {
		fprintf( stderr , "Cannot open file %s. Return value %d\n Argument should be a First Choice database file for readonly parsing\n", argv[1],fd ) ;
		exit(1) ;
	}
	
	readhead() ;
	printf("<TABLE>\n") ;
	printerr1("<CAPTION>Field names and maximum lengths</CAPTION>\n") ;
	while ( readrecord() ) ;
#ifdef FIRST2TELL
	fieldlength() ;
#endif
	printf("</TABLE>\n") ;
	return 0 ;
}
